import logging
import os
from pathlib import Path
from typing import Text, Dict, Any, Optional, List

import pytest
from _pytest.logging import LogCaptureFixture
from rasa.constants import DEFAULT_DATA_PATH
from rasa.nlu.constants import MESSAGE_INTENT_ATTRIBUTE

from rasa.nlu.training_data.formats import MarkdownReader, RasaReader
from rasa.nlu.training_data.loading import MARKDOWN, RASA

from rasax.community import config, utils
from rasax.community import config
from rasax.community.services.data_service import (
    DataService,
    training_data_dict_get_entity,
)
from rasax.community.services.domain_service import DomainService
from rasax.community.services.intent_service import INTENT_EXAMPLES_KEY
from rasax.community.services.story_service import StoryService
from rasax.community.database.data import (
    EntitySynonym,
    EntitySynonymValue,
    TrainingDataEntity,
)
from tests.unit.conftest import NLU_DATA_PATH, _user_event
from tests.unit.conftest import TEST_USER, user_dict

TEST_PROJECT = config.project_name

TEST_EXAMPLE = {
    "text": "a great example",
    "intent": "testing",
    "entities": [
        {
            "start": 0,
            "end": 1,
            "value": "indefinite",
            "entity": "article",
            "entity_synonym_id": None,
        }
    ],
    "team": "team",
    "project_id": "my_project",
}

TEST_RESULT = {
    "id": 1,
    "text": TEST_EXAMPLE["text"],
    "entities": TEST_EXAMPLE["entities"],
    "intent": TEST_EXAMPLE["intent"],
    "hash": "51601e6f1383c5bc88ff25bfaa080a20",
    "annotation": {"user": TEST_USER},
}

EXAMPLE_DATA_MD = """
## intent:check_balance
- what is my balance
- how much do I have on my [savings](source_account)
- how much do I have on my [savings account](source_account:savings)
- Could I pay in [yen](currency)?

## intent:greet
- hey
- hello

## synonym:savings
- pink pig

## regex:zipcode
- [0-9]{5}

## lookup:currencies
- Yen
- USD
- Euro
- US Dollars
- Hong Kong Dollars
- US $
- €
"""

EXAMPLE_DATA_MD_WITH_RETRIEVAL_INTENTS = """
## intent:faq/ask_howdoing
- how are you?
- how is it going?
- how's it going?

## intent:faq/inform_notunderstanding
- you don't understand
- no you're not getting me
- that's not what I said
"""

TEST_LOOKUP_TABLE = """
Berlin
Hamburg
Frankfurt
Hong Kong
Los Angeles
Malmö
""".strip()


TEST_SYNONYMS = """
## intent:specify_location
- I live in [nyc](city)
- I currently live in [nyc](city)
- I'm in [ny](city)
- I'm currently in [baires](city:buenos aires)

## synonym:new york
- nyc
- ny
- manhattan

## synonym:buenos aires
- ba
- bsas
"""


@pytest.fixture(autouse=True)
def _clear_all_synonyms(data_service: DataService):
    data_service.session.query(EntitySynonymValue).delete()
    data_service.session.query(EntitySynonym).delete()


def test_save_example(data_service: DataService):
    save_result = data_service.save_example(
        user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE
    )

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert save_result == actual[0]
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_save_example_if_it_already_exists(data_service: DataService):
    updated_example = TEST_EXAMPLE.copy()
    updated_intent = "another intent"
    updated_example["intent"] = updated_intent

    data_service.save_example(user_dict()["username"], TEST_PROJECT, updated_example)

    expected = TEST_RESULT.copy()
    expected["intent"] = updated_intent

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == expected


def test_get_by_hash(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)
    actual = data_service.get_example_by_hash(TEST_PROJECT, TEST_RESULT["hash"])

    assert actual["annotation"].pop("time") > 0
    assert actual == TEST_RESULT


def test_get_by_intent_query(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)
    intent_query = "some intent,other_intent,{}".format(TEST_EXAMPLE["intent"])

    actual, _ = data_service.get_training_data(TEST_PROJECT, intent_query=intent_query)

    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_get_with_text_query(data_service: DataService):
    text_query = "great"

    actual, _ = data_service.get_training_data(TEST_PROJECT, text_query=text_query)

    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_first_level_field_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("text", True)]
    )
    assert actual[0] == {"text": TEST_EXAMPLE["text"]}


def test_relationship_field_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("entities", True)]
    )
    assert actual[0] == {"entities": TEST_EXAMPLE["entities"]}


def test_relationship_subfield_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("entities.entity", True)]
    )
    assert actual[0] == {"entities": TEST_EXAMPLE["entities"]}


def test_get_intents(data_service: DataService):
    actual = data_service.get_intents(TEST_PROJECT)

    assert actual == [
        {"intent": TEST_RESULT["intent"], INTENT_EXAMPLES_KEY: [TEST_RESULT["hash"]]}
    ]


def test_get_intents_without_example_hashes(data_service: DataService):
    actual = data_service.get_intents(TEST_PROJECT, include_example_hashes=False)

    assert actual == [{"intent": TEST_RESULT["intent"]}]


def test_update_intent(data_service: DataService):
    new_intent = "new intent"
    data_service.update_intent(new_intent, [TEST_RESULT["hash"]], TEST_PROJECT)

    assert data_service.get_intents(TEST_PROJECT) == [
        {"intent": new_intent, INTENT_EXAMPLES_KEY: [TEST_RESULT["hash"]]}
    ]


def test_replace_example(data_service: DataService):
    updated_example = TEST_EXAMPLE.copy()
    updated_example["text"] = "another text"

    data_service.replace_example(
        user_dict(), TEST_PROJECT, updated_example, TEST_RESULT["id"], "some file"
    )

    expected = TEST_RESULT.copy()
    expected["text"] = "another text"
    expected["hash"] = "2275675a128d775f75ea2c4f3ba6be46"

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == expected


def test_get_entities(data_service: DataService):
    assert data_service.get_entities(TEST_PROJECT) == [
        {"entity": TEST_RESULT["entities"][0]["entity"]}
    ]


def test_get_warnings(data_service: DataService):
    expected = [
        {"type": "data", "min": 4, "count": 1},
        {"type": "intent", "min": 2, "count": 1, "meta": None},
        {"type": "dataPerIntent", "min": 2, "count": 1, "meta": "testing"},
    ]

    assert data_service.get_training_data_warnings(TEST_PROJECT) == expected


def test_delete_example(data_service: DataService):
    actual = data_service.delete_example(TEST_RESULT["id"])

    assert actual
    assert data_service.get_training_data(TEST_PROJECT) == ([], 0)


def test_delete_example_by_hash(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)

    assert data_service.delete_example_by_hash(TEST_PROJECT, TEST_RESULT["hash"])
    assert data_service.get_training_data(TEST_PROJECT) == ([], 0)


def test_replace_additional_training_features(
    data_service: DataService, new_project: Text
):
    # Fix project directory so it's not affected by other tests
    utils.set_project_directory("")

    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    assert data_service.get_regex_features(new_project).result == [
        {"id": 1, "name": "zipcode", "pattern": "[0-9]{5}"}
    ]
    assert data_service.get_lookup_tables(new_project) == [
        {
            "id": 1,
            "name": "currencies",
            "elements": str(Path() / DEFAULT_DATA_PATH / "currencies"),
            "number_of_elements": 7,
        }
    ]

    synonyms = data_service.get_entity_synonyms(new_project)
    assert synonyms[0]["value"] == "savings"
    assert set(synonyms[0]["synonyms"]) == {"pink pig", "savings account"}


def test_store_lookup_table(data_service: DataService, new_project: Text, tmpdir: Path):
    # Fix project directory so it's not affected by other tests
    utils.set_project_directory(str(tmpdir))

    filename = "test"
    created = data_service.save_lookup_table(filename, TEST_LOOKUP_TABLE, new_project)

    actual = data_service.get_lookup_tables(new_project)
    assert {
        "id": created["id"],
        "name": filename,
        "elements": str(tmpdir / DEFAULT_DATA_PATH / filename),
        "number_of_elements": 6,
    } in actual

    assert data_service.get_lookup_table(created["id"]) == TEST_LOOKUP_TABLE


def test_dump_lookup_table_in_local_mode(data_service: DataService, new_project: Text):
    utils.set_project_directory(".")

    nlu_file_path = os.path.join(DEFAULT_DATA_PATH, config.default_nlu_filename)
    # Insert example to have an NLU file which can be assigned as filename to the lookup
    # table
    _ = data_service.save_example(
        user_dict()["username"], new_project, TEST_EXAMPLE, filename=nlu_file_path
    )

    filename = "cars.txt"
    _ = data_service.save_lookup_table(filename, TEST_LOOKUP_TABLE, new_project)

    assert os.path.exists(os.path.join(DEFAULT_DATA_PATH, filename))
    assert os.path.exists(nlu_file_path)

    config.LOCAL_MODE = False


def test_deleting_dumped_lookup_table_in_local_mode(
    data_service: DataService, new_project: Text
):
    config.LOCAL_MODE = True
    filename = "cars.txt"
    created = data_service.save_lookup_table(filename, TEST_LOOKUP_TABLE, new_project)

    assert os.path.exists(os.path.join(DEFAULT_DATA_PATH, filename))

    data_service.delete_lookup_table(created["id"])

    assert not os.path.exists(os.path.join(DEFAULT_DATA_PATH, filename))

    config.LOCAL_MODE = False


def test_get_lookup_tables_with_elements(data_service: DataService, new_project: Text):
    filename1 = "test.txt"
    data_service.save_lookup_table(filename1, TEST_LOOKUP_TABLE, new_project)

    lookup_table2 = "Chicago\nSF\n"

    filename2 = "test2.txt"
    data_service.save_lookup_table(filename2, lookup_table2, new_project)

    actual = data_service.get_lookup_tables_with_elements(new_project)

    assert actual == [
        {
            "name": filename1,
            "elements": [
                "Berlin",
                "Hamburg",
                "Frankfurt",
                "Hong Kong",
                "Los Angeles",
                "Malmö",
            ],
        },
        {"name": filename2, "elements": ["Chicago", "SF"]},
    ]


def test_delete_lookup_table(data_service: DataService, new_project: Text):
    lookup_table = ""
    filename = "test"
    created = data_service.save_lookup_table(filename, lookup_table, new_project)

    data_service.delete_lookup_table(created["id"])

    still_existing = data_service.get_lookup_tables(new_project)
    assert created not in still_existing


def test_delete_lookup_table_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.delete_lookup_table(3456789)


def test_add_regex_feature(data_service: DataService, new_project: Text):
    training_example = {"name": "test_add_regex_feature", "pattern": ".*"}
    created = data_service.create_regex_feature(training_example, new_project)

    assert created["id"]
    actual = data_service.get_regex_feature_by_id(created["id"])
    actual.pop("id")

    assert actual == training_example


def test_add_regex_feature_with_same_pattern(
    data_service: DataService, new_project: Text
):
    training_example = {
        "name": "test_add_regex_feature_with_same_pattern",
        "pattern": ".*",
    }
    _ = data_service.create_regex_feature(training_example, new_project)

    with pytest.raises(ValueError):
        data_service.create_regex_feature(training_example, new_project)


def test_update_regex(data_service: DataService, new_project: Text):
    training_example = {
        "name": "test_add_regex_feature_with_same_pattern",
        "pattern": ".*",
    }
    created = data_service.create_regex_feature(training_example, new_project)

    updated = {"id": created["id"], "name": "so updated", "pattern": "sdadas"}

    assert data_service.update_regex_feature(created["id"], updated) == updated
    assert data_service.get_regex_feature_by_id(created["id"]) == updated


def test_update_regex_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.update_regex_feature(9929, {})


def test_delete_regex(data_service: DataService, new_project: Text):
    created = data_service.create_regex_feature(
        {"name": "test_delete_endpoint", "pattern": ".*"}, new_project
    )

    data_service.delete_regex_feature(created["id"])

    with pytest.raises(ValueError):
        data_service.get_regex_feature_by_id(created["id"])


def test_delete_regex_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.delete_regex_feature(9929)


def test_get_synonyms_entity_rasa_reader(data_service: DataService):
    # insert example data
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(TEST_PROJECT, training_data)

    # get entity synonyms and ensure RasaReader reads them correctly
    formatted_synonyms = data_service.get_entity_synonyms(TEST_PROJECT)
    assert formatted_synonyms
    training_data_json = {"rasa_nlu_data": {"entity_synonyms": formatted_synonyms}}
    training_data = RasaReader().read_from_json(training_data_json)
    assert training_data.entity_synonyms == {
        "savings account": "savings",
        "pink pig": "savings",
    }


def test_get_entity_synonyms(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Get the entity synonyms of a valid project
    synonyms = data_service.get_entity_synonyms(new_project, nlu_format=False)
    assert len(synonyms) == 1

    assert {v["value"] for v in synonyms[0]["mapped_values"]} == {
        "savings account",
        "pink pig",
    }

    # Try getting the entity synonyms of an invalid project
    synonyms = data_service.get_entity_synonyms("foobarfoobar", nlu_format=False)
    assert len(synonyms) == 0


def test_entity_synonyms_use_counts(data_service: DataService, new_project: Text):
    data_service.replace_data(new_project, TEST_SYNONYMS, MARKDOWN)

    synonyms = data_service.get_entity_synonyms(new_project, nlu_format=False)

    counts = {}
    for entry in synonyms:
        for mapped_value in entry["mapped_values"]:
            counts[mapped_value["value"]] = mapped_value["nlu_examples_count"]

    assert counts == {
        "baires": 1,
        "ba": 0,
        "bsas": 0,
        "nyc": 2,
        "ny": 1,
        "manhattan": 0,
    }


def test_create_entity_synonym(data_service: DataService, new_project: Text):
    # Create a valid entity synonym with some value mappings
    entity_synonym = data_service.create_entity_synonym(
        new_project, "United States", {"US", "USA"}
    )

    assert entity_synonym.name == "United States"
    assert {item.name for item in entity_synonym.synonym_values} == {"US", "USA"}

    # Try creating an entity synonym with no synonym values
    with pytest.raises(ValueError):
        data_service.create_entity_synonym(new_project, "foobar", set())

    # Try creating another entity synonym with the same value as another one
    # that already exists
    entity_synonym = data_service.create_entity_synonym(
        new_project, "United States", {"US", "USA"}
    )
    assert entity_synonym is None

    # Try creating an entity synonym with mapped values, with one of the mapped
    # values being identical to the entity synonym's value itself
    with pytest.raises(ValueError):
        data_service.create_entity_synonym(new_project, "foobar", {"foobar"})


def test_entity_synonym_referencing_when_adding_synonym_value(
    data_service: DataService, new_project: Text
):
    # Insert some entities which have no referenced synonym
    mobile_synonym_value = "mobile"
    entities = [
        TrainingDataEntity(
            entity="phone", value="phone", original_value=mobile_synonym_value
        ),
        TrainingDataEntity(entity="name", value="Max", original_value="Max"),
    ]
    data_service.add_all(entities)

    # Insert entity synonyms
    other_synonym_value = "other"
    created_synonym = data_service.create_entity_synonym(
        new_project, "phone synonym", {mobile_synonym_value, other_synonym_value}
    )

    # Get usage counts
    mappings_usage_counts = data_service._entity_synonym_values_use_counts(
        created_synonym
    )
    id_of_mobile_synonym_value = next(
        (
            value.id
            for value in created_synonym.synonym_values
            if value.name == mobile_synonym_value
        )
    )
    id_of_other_synonym_value = next(
        (
            value.id
            for value in created_synonym.synonym_values
            if value.name == other_synonym_value
        )
    )
    assert mappings_usage_counts == {
        id_of_mobile_synonym_value: 1,
        id_of_other_synonym_value: 0,
    }


def test_delete_synonym_value(data_service: DataService, new_project: Text):
    # Insert some entities which have no referenced synonym
    mobile_synonym_value = "mobile"
    entities = [
        TrainingDataEntity(
            entity="phone", value="phone", original_value=mobile_synonym_value
        ),
        TrainingDataEntity(entity="name", value="Max", original_value="Max"),
    ]
    data_service.add_all(entities)

    # Insert entity synonyms
    other_synonym_value = "other"
    created_synonym = data_service.create_entity_synonym(
        new_project, "phone synonym", {mobile_synonym_value, other_synonym_value}
    )

    mobile_mapping = next(
        value
        for value in created_synonym.synonym_values
        if value.name == mobile_synonym_value
    )

    # Delete mapping
    assert data_service.delete_entity_synonym_mapped_value(
        new_project, created_synonym.id, mobile_mapping.id
    )

    # Assert reference was deleted from entity
    assert entities[0].entity_synonym_id is None


def test_get_entity_synonym(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Get a valid entity synonym
    entity_synonym = data_service.get_entity_synonym(new_project, 1)

    assert entity_synonym["id"] == 1
    assert {v["value"] for v in entity_synonym["mapped_values"]} == {
        "savings account",
        "pink pig",
    }

    # Try getting an invalid entity synonym from a valid project
    assert data_service.get_entity_synonym(new_project, 9999) is None

    # Try getting an entity synonym from an invalid project
    assert data_service.get_entity_synonym("foobarfoobar", 1) is None


def test_get_entity_synonym_internal(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Get a valid entity synonym
    entity_synonym = data_service._get_entity_synonym(new_project, 1)

    assert entity_synonym.id == 1
    assert {v.name for v in entity_synonym.synonym_values} == {
        "savings account",
        "pink pig",
    }

    # Try getting an invalid entity synonym from a valid project
    assert data_service._get_entity_synonym(new_project, 9999) is None

    # Try getting an entity synonym from an invalid project
    assert data_service._get_entity_synonym("foobarfoobar", 1) is None


def test_add_entity_synonym_values(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Add a new mapping to an entity synonym
    values = data_service.add_entity_synonym_mapped_values(new_project, 1, {"money"})
    assert [item["value"] for item in values] == [
        "money"
    ]  # Added mappings are returned

    # Retrieve the entity synonym again and check added values there
    entity_synonym = data_service.get_entity_synonym(new_project, 1)
    assert {item["value"] for item in entity_synonym["mapped_values"]} == {
        "pink pig",
        "savings account",
        "money",
    }

    # Try adding mapped values to an invalid entity synonym
    with pytest.raises(ValueError):
        values = data_service.add_entity_synonym_mapped_values(
            new_project, 9999, {"foobar"}
        )


@pytest.mark.parametrize(
    "synonyms, expected_synonyms",
    [
        (["money", "money", "money", "coins"], ["coins", "money"]),
        (["savings", "coins"], ["coins"]),
    ],
)
def test_add_entity_synonym_values_warnings(
    data_service: DataService,
    new_project: Text,
    synonyms: List[Text],
    expected_synonyms: List[Text],
    caplog: LogCaptureFixture,
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    with caplog.at_level(logging.WARNING):
        values = data_service.add_entity_synonym_mapped_values(new_project, 1, synonyms)

    assert sorted([item["value"] for item in values]) == sorted(expected_synonyms)
    assert len(caplog.records) == 1


def test_add_entity_synonym_values_with_previously_mapped_values(
    data_service: DataService, new_project: Text, caplog: LogCaptureFixture
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    data_service.add_entity_synonym_mapped_values(new_project, 1, ["money"])

    with caplog.at_level(logging.WARNING):
        values = data_service.add_entity_synonym_mapped_values(
            new_project, 1, ["money"]
        )

    assert [item["value"] for item in values] == []
    assert len(caplog.records) == 1


def test_delete_entity_synonym(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Try to delete an EntitySynonym that does not exist
    deleted = data_service.delete_entity_synonym(new_project, 9999)
    assert not deleted

    # Delete an existing EntitySynonym
    deleted = data_service.delete_entity_synonym(new_project, 1)
    assert deleted
    entity_synonym = data_service._get_entity_synonym(new_project, 1)
    assert entity_synonym is None


def test_update_entity_synonym(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Update an entity synonym's value (.name property)
    entity_synonym = data_service.update_entity_synonym(new_project, 1, "bank")
    assert entity_synonym.name == "bank"  # Updated entity synonym is returned

    # Retrieve the same entity synonym again and check the change there
    entity_synonym = data_service._get_entity_synonym(new_project, 1)
    assert entity_synonym.name == "bank"

    # Try updating an invalid entity synonym
    with pytest.raises(ValueError):
        data_service.update_entity_synonym(new_project, 9999, "foobar")


def test_update_entity_synonym_repeated(data_service: DataService, new_project: Text):
    training_data = MarkdownReader().reads(TEST_SYNONYMS)
    data_service.replace_additional_training_features(new_project, training_data)

    entity_synonym = data_service._get_entity_synonym(new_project, 1)

    # Depending on which entity synonym we got, try changing its value to the
    # value of the other one
    if entity_synonym.name == "new york":
        updated = data_service.update_entity_synonym(new_project, 1, "buenos aires")
    elif entity_synonym.name == "buenos aires":
        updated = data_service.update_entity_synonym(new_project, 1, "new york")
    else:
        raise ValueError

    # EntitySynonym should not have been updated because there's another
    # EntitySynonym with the same name
    assert updated is None


def test_delete_entity_synonym_mapped_value(
    data_service: DataService, new_project: Text
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Try deleting an invalid value mapping of an invalid entity synonym
    with pytest.raises(ValueError):
        data_service.delete_entity_synonym_mapped_value(new_project, 9999, 9999)

    # Try deleting an invalid value mapping of a valid entity synonym
    assert not data_service.delete_entity_synonym_mapped_value(new_project, 1, 9999)

    # Delete a value mapping of a valid entity synonym
    assert data_service.delete_entity_synonym_mapped_value(new_project, 1, 1)


def test_delete_last_entity_synonym_mapped_value(
    data_service: DataService, new_project: Text
):
    # Insert entity synonyms
    created_synonym = data_service.create_entity_synonym(
        new_project, "phone synonym", {"mapping"}
    )

    assert len(created_synonym.synonym_values) == 1
    assert data_service.delete_entity_synonym_mapped_value(
        new_project, created_synonym.id, created_synonym.synonym_values[0].id
    )

    # Assert synonym was also deleted because last mapped value for this synonym was
    # deleted
    assert data_service.get_entity_synonym(new_project, created_synonym.id) is None


def test_get_entity_synonym_value_internal(
    data_service: DataService, new_project: Text
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Try getting a value mapping of an invalid entity synonym
    with pytest.raises(ValueError):
        data_service._get_entity_synonym_value(new_project, 9999, 9999)

    # Try getting an invalid value mapping (by ID) of a valid entity synonym
    entity_synonym_value = data_service._get_entity_synonym_value(new_project, 1, 9999)
    assert entity_synonym_value is None

    # Get both value mappings (by ID) for entity synonym with ID 1
    entity_synonym_value_1 = data_service._get_entity_synonym_value(new_project, 1, 1)
    entity_synonym_value_2 = data_service._get_entity_synonym_value(new_project, 1, 2)
    assert {entity_synonym_value_1.name, entity_synonym_value_2.name} == {
        "pink pig",
        "savings account",
    }


def test_get_entity_synonym_value_by_name_internal(
    data_service: DataService, new_project: Text
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Try getting a value mapping of an invalid entity synonym
    with pytest.raises(ValueError):
        data_service._get_entity_synonym_value_by_name(new_project, 9999, "foobar")

    # Try getting an invalid value mapping (by name) of a valid entity synonym
    entity_synonym_value = data_service._get_entity_synonym_value_by_name(
        new_project, 1, "foobar"
    )
    assert entity_synonym_value is None

    # Get both value mappings (by name) for entity synonym with ID 1
    entity_synonym_value_1 = data_service._get_entity_synonym_value_by_name(
        new_project, 1, "pink pig"
    )
    entity_synonym_value_2 = data_service._get_entity_synonym_value_by_name(
        new_project, 1, "savings account"
    )
    assert {entity_synonym_value_1.name, entity_synonym_value_2.name} == {
        "pink pig",
        "savings account",
    }


def test_get_synonym_id_for_mapped_value(data_service: DataService, new_project: Text):
    created_synonym = data_service.create_entity_synonym(new_project, "house", {"hut"})

    assert (
        data_service.get_synonym_by_mapped_value("hut", new_project)
        == created_synonym.as_dict()
    )


def test_get_synonym_id_for_mapped_value_if_no_matching_mapped_value(
    data_service: DataService, new_project: Text
):
    data_service.create_entity_synonym(new_project, "house", {"hut"})

    assert data_service.get_synonym_by_mapped_value("hut2", new_project) is None


def test_update_entity_synonym_mapped_value(
    data_service: DataService, new_project: Text
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    # Try updating a value mapping for an invalid entity synonym
    with pytest.raises(ValueError):
        data_service.update_entity_synonym_mapped_value(new_project, 9999, 1, "")

    # Try updating an invalid value mapping for an existing entity synonym
    with pytest.raises(ValueError):
        data_service.update_entity_synonym_mapped_value(new_project, 1, 9999, "")

    # Try adding a mapping that already exists
    entity_synonym_value = data_service.update_entity_synonym_mapped_value(
        new_project, 1, 1, "savings account"
    )
    assert entity_synonym_value is None

    # Update a valid mapping
    entity_synonym_value = data_service.update_entity_synonym_mapped_value(
        new_project, 1, 1, "bank"
    )
    assert entity_synonym_value.name == "bank"

    # Retrieve the value again to check that it is updated
    entity_synonym = data_service.get_entity_synonym(new_project, 1)
    assert len(entity_synonym["mapped_values"]) == 2

    # Try setting a mapping value to the value of its entity synonym
    with pytest.raises(ValueError):
        data_service.update_entity_synonym_mapped_value(new_project, 1, 1, "savings")


def test_update_entity_synonym_mapped_value_with_mapped_entities(
    data_service: DataService, new_project: Text
):
    # Insert some entities which have no referenced synonym
    entities = [
        TrainingDataEntity(entity="house", value="house", original_value="hut"),
        TrainingDataEntity(entity="house", value="building", original_value="building"),
    ]
    data_service.add_all(entities)

    created_synonym = data_service.create_entity_synonym(new_project, "house", {"hut"})

    mapped_synonym_value_id = created_synonym.synonym_values[0].id

    mappings_usage_counts = data_service._entity_synonym_values_use_counts(
        created_synonym
    )
    assert mappings_usage_counts[mapped_synonym_value_id] == 1

    data_service.update_entity_synonym_mapped_value(
        new_project, created_synonym.id, mapped_synonym_value_id, "something else"
    )

    mappings_usage_counts = data_service._entity_synonym_values_use_counts(
        created_synonym
    )
    assert mappings_usage_counts[mapped_synonym_value_id] == 0

    data_service.update_entity_synonym_mapped_value(
        new_project, created_synonym.id, mapped_synonym_value_id, "building"
    )

    mappings_usage_counts = data_service._entity_synonym_values_use_counts(
        created_synonym
    )
    assert mappings_usage_counts[mapped_synonym_value_id] == 1


def test_domain_items_from_nlu_data(
    data_service: DataService, domain_service: DomainService
):
    from tests.unit.services.test_domain_service import TEST_DOMAIN

    # ensure there is a domain
    domain_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=False
    )

    # add an example
    data_service.save_example(
        TEST_USER, TEST_PROJECT, TEST_EXAMPLE, add_example_items_to_domain=False
    )
    data_service.add_all_intents_and_entities_to_domain(TEST_PROJECT)

    domain_intents = domain_service.get_intents_from_domain(TEST_PROJECT)
    domain_entities = domain_service.get_entities_from_domain(TEST_PROJECT)

    # ensure domain contains example's entity and intent
    assert TEST_EXAMPLE["intent"] in domain_intents
    assert TEST_EXAMPLE["entities"][0]["entity"] in domain_entities


async def test_domain_items_from_story(
    story_service: StoryService, domain_service: DomainService
):
    from tests.unit.services.test_domain_service import TEST_DOMAIN

    # ensure there is a domain
    domain_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=False
    )

    # add a story
    intent = "some_intent1"
    action = "some_action1"
    story = """
    ## story1
    * {}
      - {}
    """.format(
        intent, action
    )
    await story_service.save_stories(
        story,
        "team",
        username=TEST_USER,
        dump_stories=False,
        add_story_items_to_domain=True,
    )

    domain_actions = domain_service.get_actions_from_domain(TEST_PROJECT)
    domain_intents = domain_service.get_intents_from_domain(TEST_PROJECT)

    # ensure domain contain story action and intent
    assert action in domain_actions
    assert intent in domain_intents


def test_inject_data_from_file(data_service: DataService, new_project: Text):
    assert not data_service.get_training_data(new_project)[0]
    expected = MarkdownReader().read(NLU_DATA_PATH)

    # insert data from file
    data_service.save_bulk_data_from_files([NLU_DATA_PATH], new_project, TEST_USER)
    assert data_service.get_training_data(new_project)[0]

    actual = data_service.get_nlu_training_data_object(project_id=new_project)

    # use markdown to test equality since json can contain ids from the database
    assert actual.training_examples == expected.training_examples

    lookup_tables = data_service.get_lookup_tables(new_project)
    assert len(lookup_tables) == 1


def test_inject_data_with_invalid_file_ending(
    data_service: DataService, tmp_path: Path
):
    file = tmp_path / "my_text.txt"
    file.write_text("test test test test")
    inserted = data_service.save_bulk_data_from_files([str(file)])
    assert inserted.is_empty()


def test_entity_extractor_persistence(data_service: DataService):
    extractor = "some_extractor"
    test_example_with_extractor = TEST_EXAMPLE.copy()
    test_example_with_extractor["entities"][0]["extractor"] = extractor
    # modify text so it's saved as a new example
    new_text = "new_text"
    test_example_with_extractor["text"] = new_text

    # save example
    save_result = data_service.save_example(
        user_dict()["username"], TEST_PROJECT, test_example_with_extractor
    )

    # make sure extractor was saved
    actual, number_examples = data_service.get_training_data(
        TEST_PROJECT, text_query=new_text
    )
    assert number_examples == 1
    assert save_result == actual[0]
    assert actual[0]["entities"][0]["extractor"] == extractor


@pytest.mark.parametrize(
    "training_data, format",
    [
        (MarkdownReader().reads(EXAMPLE_DATA_MD).as_markdown(), MARKDOWN),
        (MarkdownReader().reads(EXAMPLE_DATA_MD).as_json(), RASA),
    ],
)
def test_replace_data(
    data_service: DataService, new_project: Text, training_data: Text, format: Text
):
    data_service.replace_data(new_project, training_data, format)

    assert data_service.get_training_data(new_project)


def test_save_bulk_data_with_unknown_format(data_service: DataService):
    assert not data_service.save_bulk_data("project", "some data", "unknown format")


def test_get_all_filenames(data_service: DataService, new_project: Text):
    filenames = {"one", "two"}
    for file in filenames:
        example = TEST_EXAMPLE.copy()
        example["text"] = file
        _ = data_service.save_example(
            user_dict()["username"], new_project, example, filename=file
        )

    assert set(data_service.get_all_filenames(new_project)) == filenames

    assert all(
        [
            len(data_service.get_training_data_for_filename(new_project, f)) == 1
            for f in filenames
        ]
    )


def test_assign_filename_based_on_intent(data_service: DataService, new_project: Text):
    # Insert example to make sure the `oldest` file is a different one than the one we
    # actually we want to test
    oldest_file = "very old"
    data_service.save_example(
        user_dict()["username"], new_project, TEST_EXAMPLE.copy(), filename=oldest_file
    )

    intent = "make intents great again"

    # Insert example which will specify the file for the intent
    file_name = "filename.txt"
    example = TEST_EXAMPLE.copy()
    example["text"] = "💪🏽"
    example["intent"] = intent
    expected1 = data_service.save_example(
        user_dict()["username"], new_project, example, filename=file_name
    )

    # This example should get the same file assigned as the previous one since they
    # have the same intent
    example["text"] = "👀"
    # This example should get the other file assigned.
    expected2 = data_service.save_example(user_dict()["username"], new_project, example)

    assert data_service.get_training_data_for_filename(new_project, file_name) == [
        expected1,
        expected2,
    ]


def test_assign_filename_based_on_intent_if_no_example_with_intent_exists(
    data_service: DataService, new_project: Text
):
    # Insert example to make sure their is an `oldest` file.
    oldest_file = "very old"
    old_example = data_service.save_example(
        user_dict()["username"], new_project, TEST_EXAMPLE.copy(), filename=oldest_file
    )

    # Insert example with a new intent which is not yet in the examples
    example = TEST_EXAMPLE.copy()
    example["text"] = "💪🏽"
    example["intent"] = "🛠"
    expected1 = data_service.save_example(user_dict()["username"], new_project, example)

    assert data_service.get_training_data_for_filename(new_project, oldest_file) == [
        old_example,
        expected1,
    ]


def test_save_user_event_as_example(data_service: DataService, new_project: Text):
    test_text = "Ready to launch"
    test_intent = "test_save_user_event_as_example"
    event = _user_event(
        "sender", intent=test_intent, text=test_text, entities=TEST_EXAMPLE["entities"]
    )
    data_service.save_user_event_as_example(user_dict(), new_project, event)

    actual, number_of_results = data_service.get_training_data(new_project)

    assert number_of_results == 1
    assert actual[0]["text"] == test_text
    assert actual[0]["intent"] == test_intent
    assert actual[0]["entities"] == TEST_EXAMPLE["entities"]


def test_dump_data(data_service: DataService, new_project: Text):
    filenames = {"dir/one.md", "dir/two.md"}
    for file in filenames:
        example = TEST_EXAMPLE.copy()
        example["text"] = file
        _ = data_service.save_example(
            user_dict()["username"], new_project, example, filename=file
        )

    data_service._dump_data(new_project)

    assert all([os.path.exists(f) for f in filenames])


@pytest.mark.parametrize(
    "training_example, start, end, expected",
    [({"text": "I live in nyc"}, 10, 13, "nyc"), ({"text": "hello"}, None, None, None)],
)
def test_training_data_dict_get_entity(
    training_example: Dict[Text, Any],
    start: Optional[int],
    end: Optional[int],
    expected: Optional[Text],
):
    assert expected == training_data_dict_get_entity(training_example, start, end)


def test_retrieval_intent_merging(data_service: DataService):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD_WITH_RETRIEVAL_INTENTS)

    example_intent = "faq"

    # training data contains non-merged intents
    assert all(
        example.get(MESSAGE_INTENT_ATTRIBUTE) == example_intent
        for example in training_data.training_examples
    )

    # noinspection PyProtectedMember
    data_service._combine_response_keys_and_intents(training_data)

    # ensure training data now contains merged intents, i.e. "faq/<something>"
    assert all(
        example.get(MESSAGE_INTENT_ATTRIBUTE).startswith(f"{example_intent}/")
        for example in training_data.training_examples
    )


@pytest.mark.parametrize(
    "markdown_data,contains_retrieval_intents",
    [(EXAMPLE_DATA_MD_WITH_RETRIEVAL_INTENTS, True), (EXAMPLE_DATA_MD, False)],
)
def test_training_data_contains_retrieval_intents(
    data_service: DataService, markdown_data: Text, contains_retrieval_intents: bool
):
    training_data = MarkdownReader().reads(markdown_data)
    assert (
        data_service.training_data_contains_retrieval_intents(training_data)
        == contains_retrieval_intents
    )
